/*
 *  xtiger port on PSP 
 *
 *  Copyright (C) 2006 Ludovic Jacomme (ludovic.jacomme@gmail.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <fcntl.h>

#include <pspctrl.h>
#include <pspkernel.h>
#include <pspdebug.h>

#include "global.h"
#include "sysdeps.h"
#include "config.h"
#include "options.h"
#include "newcpu.h"
#include "keyboard.h"
#include "specific.h"
#include "cmdinterface.h"
#include "debug.h"
#include "globinfo.h"

#include "psp_kbd.h"
#include "psp_pg.h"
#include "psp_init.h"
#include "psp_specific.h"

  long ti92_xor_color[2][3] = { 
  {
    PG_RGB(0xFe,0xFe,0xFe), //  0xFF, 0xFF, 0xFF
    PG_RGB(0xFF,0xFF,0xFF), //  0xFF, 0xFF, 0xFF
    PG_RGB(0xFF,0x0F,0x00)  //  0xFF, 0xfF, 0x00
  },
  {
    PG_RGB(0xFF,0xFF,0xFF), //  0xFF, 0xFF, 0xFF
    PG_RGB(0xFF,0xFF,0xFF), //  0xFF, 0xFF, 0xFF
    PG_RGB(0xFF,0xFF,0xFF)  //  0xFF, 0xfF, 0x00
  } };

  static ti92_button_t ti92_buttons[2][  TI92_LAST_BUTTON ] = {
  {
    { "Q",       TIKEY_Q           ,   2, 161, 28, 19, 0 }, 
    { "W",       TIKEY_W           ,  36, 161, 29, 19, 0 }, 
    { "E",       TIKEY_E           ,  70, 161, 28, 19, 0 }, 
    { "R",       TIKEY_R           , 104, 161, 29, 19, 0 }, 
    { "T",       TIKEY_T           , 138, 161, 29, 19, 0 }, 
    { "Y",       TIKEY_Y           , 172, 161, 29, 19, 0 }, 
    { "U",       TIKEY_U           , 206, 161, 29, 19, 0 }, 
    { "I",       TIKEY_I           , 240, 161, 29, 19, 0 }, 
    { "O",       TIKEY_O           , 274, 161, 29, 19, 0 }, 
    { "P",       TIKEY_P           , 309, 161, 28, 19, 0 }, 
    { "7",       TIKEY_7           , 339, 161, 34, 20, 1 }, 
    { "8",       TIKEY_8           , 374, 161, 34, 20, 1 }, 
    { "9",       TIKEY_9           , 409, 161, 34, 20, 1 }, 
    { "*",       TIKEY_MULTIPLY    , 445, 161, 33, 20, 0 }, 
                                   
    { "A",       TIKEY_A           ,  19, 190, 29, 19, 0 }, 
    { "S",       TIKEY_S           ,  53, 190, 28, 19, 0 }, 
    { "D",       TIKEY_D           ,  87, 190, 29, 19, 0 }, 
    { "F",       TIKEY_F           , 121, 190, 29, 19, 0 }, 
    { "G",       TIKEY_G           , 155, 190, 29, 19, 0 }, 
    { "H",       TIKEY_H           , 189, 190, 29, 19, 0 }, 
    { "J",       TIKEY_J           , 224, 190, 28, 19, 0 }, 
    { "K",       TIKEY_K           , 257, 190, 29, 19, 0 }, 
    { "L",       TIKEY_L           , 292, 190, 29, 19, 0 }, 
    { "4",       TIKEY_4           , 339, 190, 34, 20, 1 }, 
    { "5",       TIKEY_5           , 374, 190, 34, 20, 1 }, 
    { "6",       TIKEY_6           , 409, 190, 34, 20, 1 }, 
    { "-",       TIKEY_MINUS       , 445, 190, 33, 19, 0 }, 
                                   
    { "SHIFT",   TIKEY_SHIFT       ,   2, 219, 28, 19, 0 }, 
    { "Z",       TIKEY_Z           ,  36, 219, 29, 19, 0 }, 
    { "X",       TIKEY_X           ,  70, 219, 28, 19, 0 }, 
    { "C",       TIKEY_C           , 104, 219, 29, 19, 0 }, 
    { "V",       TIKEY_V           , 138, 219, 28, 19, 0 }, 
    { "B",       TIKEY_B           , 172, 219, 29, 19, 0 }, 
    { "N",       TIKEY_N           , 206, 219, 29, 19, 0 }, 
    { "M",       TIKEY_M           , 240, 219, 29, 19, 0 }, 
    { "THETA",   TIKEY_THETA       , 274, 219, 29, 19, 0 },
    { "ENTER",   TIKEY_ENTER1      , 309, 219, 28, 46, 2 }, 
    { "1",       TIKEY_1           , 339, 219, 34, 20, 1 }, 
    { "2",       TIKEY_2           , 374, 219, 34, 20, 1 }, 
    { "3",       TIKEY_3           , 409, 219, 34, 20, 1 }, 
    { "*",       TIKEY_PLUS        , 445, 219, 33, 19, 0 }, 
                                   
    { "ON",      TIKEY_ON          ,   2, 248, 28, 19, 0 }, 
    { "DIAMOND", TIKEY_DIAMOND     ,  36, 248, 29, 19, 0 },
    { "2ND",     TIKEY_2ND         ,  70, 248, 28, 19, 0 }, 
    { "STORE",   TIKEY_STORE       , 104, 248, 29, 19, 0 }, 
    { "SPACE",   TIKEY_SPACE       , 138, 248, 97, 19, 0 }, 
    { "=",       TIKEY_EQUALS      , 241, 248, 29, 19, 0 }, 
    { "BKSP",    TIKEY_BACKSPACE   , 274, 248, 29, 19, 0 }, 
    { "ENTER",   TIKEY_ENTER1      , 309, 219, 28, 46, 2 }, 
    { "0",       TIKEY_0           , 339, 248, 34, 20, 1 },
    { ".",       TIKEY_PERIOD      , 374, 248, 34, 20, 1 }, 
    { "(-)",     TIKEY_NEGATE      , 409, 248, 34, 20, 1 }, 
    { "ENTER",   TIKEY_ENTER2      , 445, 248, 33, 19, 2 }, 

    { "2ND",     TIKEY_2ND         , 339,   1, 34, 19, 0 },

    { "MODE",    TIKEY_MODE        , 339,  22, 33, 19, 0 },

    { "CLEAR",   TIKEY_CLEAR       , 339,  44, 33, 19, 0 },

    { "LN",      TIKEY_LN          , 339,  74, 34, 19, 0 },
    { "ESCAPE",  TIKEY_ESCAPE      , 374,  74, 34, 19, 2 },
    { "APPS" ,   TIKEY_APPS        , 409,  74, 34, 19, 2 },
    { "ENTER",   TIKEY_ENTER2      , 445,  74, 33, 19, 2 },

    { "SIN",     TIKEY_SIN         , 339, 103, 34, 19, 0 },
    { "COS",     TIKEY_COS         , 374, 103, 34, 19, 0 },
    { "TAN",     TIKEY_TAN         , 409, 103, 34, 19, 0 },
    { "^",       TIKEY_POWER       , 445, 103, 33, 19, 0 },

    { "(",       TIKEY_PALEFT      , 339, 132, 34, 19, 0 },
    { ")",       TIKEY_PARIGHT     , 374, 132, 34, 19, 0 },
    { ",",       TIKEY_COMMA       , 409, 132, 34, 19, 0 },
    { "/",       TIKEY_DIVIDE      , 445, 132, 33, 19, 0 },

    /* Do Not display */
                                   
    { "LEFT",    TIKEY_LEFT        ,   0,   0,  0,  0, 0 }, 
    { "DOWN",    TIKEY_DOWN        ,   0,   0,  0,  0, 0 }, 
    { "UP",      TIKEY_UP          ,   0,   0,  0,  0, 0 }, 
    { "RIGHT",   TIKEY_RIGHT       ,   0,   0,  0,  0, 0 }, 

    { "F1",      TIKEY_F1          ,   0,   0,  0,  0, 0 }, 
    { "F2",      TIKEY_F2          ,   0,   0,  0,  0, 0 }, 
    { "F3",      TIKEY_F3          ,   0,   0,  0,  0, 0 }, 
    { "F4",      TIKEY_F4          ,   0,   0,  0,  0, 0 }, 
    { "F5",      TIKEY_F5          ,   0,   0,  0,  0, 0 }, 
    { "F6",      TIKEY_F6          ,   0,   0,  0,  0, 0 }, 
    { "F7",      TIKEY_F7          ,   0,   0,  0,  0, 0 }, 
    { "F8",      TIKEY_F8          ,   0,   0,  0,  0, 0 }, 

    { "HAND",    TIKEY_HAND        ,   0,   0,  0,  0, 0 }, 

  },
  {
    { "Q",       TIKEY_Q           ,   2, 168, 26, 17, 0 }, 
    { "W",       TIKEY_W           ,  36, 168, 26, 17, 0 }, 
    { "E",       TIKEY_E           ,  71, 168, 26, 17, 0 }, 
    { "R",       TIKEY_R           , 107, 168, 26, 17, 0 }, 
    { "T",       TIKEY_T           , 143, 168, 26, 17, 0 }, 
    { "Y",       TIKEY_Y           , 180, 168, 26, 17, 0 }, 
    { "U",       TIKEY_U           , 217, 168, 26, 17, 0 }, 
    { "I",       TIKEY_I           , 253, 168, 26, 17, 0 }, 
    { "O",       TIKEY_O           , 290, 168, 26, 17, 0 }, 
    { "P",       TIKEY_P           , 326, 168, 26, 17, 0 }, 
    { "7",       TIKEY_7           , 357, 167, 29, 19, 1 }, 
    { "8",       TIKEY_8           , 388, 167, 29, 19, 1 }, 
    { "9",       TIKEY_9           , 418, 167, 29, 19, 1 }, 
    { "*",       TIKEY_MULTIPLY    , 450, 168, 26, 17, 0 }, 
                                   
    { "A",       TIKEY_A           ,  16, 196, 26, 17, 0 }, 
    { "S",       TIKEY_S           ,  53, 196, 26, 17, 0 }, 
    { "D",       TIKEY_D           ,  89, 196, 26, 17, 0 }, 
    { "F",       TIKEY_F           , 126, 196, 26, 17, 0 }, 
    { "G",       TIKEY_G           , 162, 196, 26, 17, 0 }, 
    { "H",       TIKEY_H           , 198, 196, 26, 17, 0 }, 
    { "J",       TIKEY_J           , 235, 196, 26, 17, 0 }, 
    { "K",       TIKEY_K           , 271, 196, 26, 17, 0 }, 
    { "L",       TIKEY_L           , 308, 196, 26, 17, 0 }, 
    { "4",       TIKEY_4           , 357, 195, 29, 19, 1 }, 
    { "5",       TIKEY_5           , 388, 195, 29, 19, 1 }, 
    { "6",       TIKEY_6           , 418, 195, 29, 19, 1 }, 
    { "-",       TIKEY_MINUS       , 450, 196, 26, 17, 0 }, 
                                   
    { "SHIFT",   TIKEY_SHIFT       ,   2, 224, 26, 17, 0 }, 
    { "Z",       TIKEY_Z           ,  35, 224, 26, 17, 0 }, 
    { "X",       TIKEY_X           ,  71, 224, 26, 17, 0 }, 
    { "C",       TIKEY_C           , 107, 224, 26, 17, 0 }, 
    { "V",       TIKEY_V           , 143, 224, 26, 17, 0 }, 
    { "B",       TIKEY_B           , 180, 224, 26, 17, 0 }, 
    { "N",       TIKEY_N           , 217, 224, 26, 17, 0 }, 
    { "M",       TIKEY_M           , 253, 224, 26, 17, 0 }, 
    { "THETA",   TIKEY_THETA       , 290, 224, 26, 17, 0 },
    { "ENTER",   TIKEY_ENTER1      , 328, 241, 25,  9, 2 }, 
    { "1",       TIKEY_1           , 357, 223, 29, 19, 1 }, 
    { "2",       TIKEY_2           , 388, 223, 29, 19, 1 }, 
    { "3",       TIKEY_3           , 418, 223, 29, 19, 1 }, 
    { "*",       TIKEY_PLUS        , 450, 224, 26, 17, 0 }, 
                                   
    { "ON",      TIKEY_ON          ,   2, 252, 26, 17, 0 }, 
    { "DIAMOND", TIKEY_DIAMOND     ,  35, 252, 26, 17, 0 },
    { "2ND",     TIKEY_2ND         ,  71, 252, 26, 17, 0 }, 
    { "STORE",   TIKEY_STORE       , 107, 252, 26, 17, 0 }, 
    { "SPACE",   TIKEY_SPACE       , 143, 252,101, 17, 0 }, 
    { "=",       TIKEY_EQUALS      , 253, 252, 26, 17, 0 }, 
    { "BKSP",    TIKEY_BACKSPACE   , 290, 252, 26, 17, 0 }, 
    { "ENTER",   TIKEY_ENTER1      , 328, 241, 25,  9, 2 }, 
    { "0",       TIKEY_0           , 357, 251, 29, 19, 1 },
    { ".",       TIKEY_PERIOD      , 388, 251, 29, 19, 1 }, 
    { "(-)",     TIKEY_NEGATE      , 418, 251, 29, 19, 1 }, 
    { "ENTER",   TIKEY_ENTER2      , 451, 260, 25,  8, 2 }, 

    { "2ND",     TIKEY_2ND         , 358,   2, 26, 21, 0 },
    { "ESCAPE",  TIKEY_ESCAPE      , 391,   9, 20, 10, 2 },

    { "MODE",    TIKEY_MODE        , 358,  31, 26, 17, 0 },
    { "ESCAPE",  TIKEY_ESCAPE      , 391,   9, 20, 10, 2 },

    { "CLEAR",   TIKEY_CLEAR       , 358,  56, 26, 17, 0 },
    { "APPS" ,   TIKEY_APPS        , 391,  63, 23,  8, 2 },

    { "LN",      TIKEY_LN          , 358,  84, 26, 17, 0 },
    { "APPS" ,   TIKEY_APPS        , 391,  63, 23,  8, 2 },
    { "ENTER",   TIKEY_ENTER2      , 433,  88, 29,  9, 2 },

    { "SIN",     TIKEY_SIN         , 358, 112, 26, 17, 0 },
    { "COS",     TIKEY_COS         , 389, 112, 26, 17, 0 },
    { "TAN",     TIKEY_TAN         , 419, 112, 26, 17, 0 },
    { "^",       TIKEY_POWER       , 450, 112, 26, 17, 0 },

    { "(",       TIKEY_PALEFT      , 358, 140, 26, 17, 0 },
    { ")",       TIKEY_PARIGHT     , 389, 140, 26, 17, 0 },
    { ",",       TIKEY_COMMA       , 419, 140, 26, 17, 0 },
    { "/",       TIKEY_DIVIDE      , 450, 140, 26, 17, 0 },

    /* Do Not display */
                                   
    { "LEFT",    TIKEY_LEFT        ,   0,   0,  0,  0, 0 }, 
    { "DOWN",    TIKEY_DOWN        ,   0,   0,  0,  0, 0 }, 
    { "UP",      TIKEY_UP          ,   0,   0,  0,  0, 0 }, 
    { "RIGHT",   TIKEY_RIGHT       ,   0,   0,  0,  0, 0 }, 

    { "F1",      TIKEY_F1          ,   0,   0,  0,  0, 0 }, 
    { "F2",      TIKEY_F2          ,   0,   0,  0,  0, 0 }, 
    { "F3",      TIKEY_F3          ,   0,   0,  0,  0, 0 }, 
    { "F4",      TIKEY_F4          ,   0,   0,  0,  0, 0 }, 
    { "F5",      TIKEY_F5          ,   0,   0,  0,  0, 0 }, 
    { "F6",      TIKEY_F6          ,   0,   0,  0,  0, 0 }, 
    { "F7",      TIKEY_F7          ,   0,   0,  0,  0, 0 }, 
    { "F8",      TIKEY_F8          ,   0,   0,  0,  0, 0 }, 

    { "HAND",    TIKEY_HAND        ,   0,   0,  0,  0, 0 }, 

  } };

int
key_event(int b, int button_pressed)
{
  int ti_key;

  ti_key = ti92_buttons[b].code;
  if (button_pressed) {
#ifdef DEBUG_BUTTONS
    pspDebugPrintf( "Button pressed %d\n", ti_key );
#endif
    keyStates[ti_key] = 1;
    if(keyStates[OPT_QUIT]) 
      specialflags |= SPCFLAG_BRK;
    else if(keyStates[OPT_LOADFILE]) {
      keyStates[OPT_LOADFILE] = 0;
      screen_off();
      enter_command();
    }

  } else {
#ifdef DEBUG_BUTTONS
    pspDebugPrintf( "Button released %d\n", ti_key );
#endif
    keyStates[ti_key] = 0;
  }
  return 0;
}

int
ti92_decode_key(int psp_b, int button_pressed)
{
  int wake = 0;
  switch ( psp_b )  {
    case KBD_LEFT     : wake = 1;
                        key_event( TI92_BUTTON_LEFT, button_pressed);
    break;
    case KBD_RIGHT    : wake = 1;
                        key_event( TI92_BUTTON_RIGHT, button_pressed);
    break;
    case KBD_UP       : wake = 1;
                        key_event( TI92_BUTTON_UP   , button_pressed);
    break;
    case KBD_DOWN     : wake = 1;
                        key_event( TI92_BUTTON_DOWN, button_pressed);
    break;
    case KBD_CROSS    : wake = 1;
                        key_event( TI92_BUTTON_ENTER_1, button_pressed);
    break;
    case KBD_SQUARE   : wake = 1;
                        key_event( TI92_BUTTON_BACKSPACE, button_pressed);
    break;
    case KBD_CIRCLE   : wake = 1;
                        key_event( ti92_cursor_button, button_pressed);
    break;
    case KBD_TRIANGLE : wake = 1;
                        key_event( TI92_BUTTON_ESCAPE, button_pressed);
    break;
    case KBD_START    : wake = 1;
                        key_event( TI92_BUTTON_F1, button_pressed);
    break;
    case KBD_SELECT   : wake = 1;
                        key_event( TI92_BUTTON_DIAMOND, button_pressed);
    break;
    case KBD_LTRIGGER : wake = 1;
                        key_event( TI92_BUTTON_HAND, button_pressed);
    break;
    case KBD_RTRIGGER : wake = 1;
                        key_event( TI92_BUTTON_2ND_1, button_pressed);
    break;
  }
  return 0;
}

static void
ti92_show_cursor_button(void)
{
  int b = ti92_cursor_button;
  if (! ti92_cursor_button_on) {
    ti92_cursor_button_on = 1;
    pgFillRectangle( ti92_buttons[b].x, 
                     ti92_buttons[b].y,
                     ti92_buttons[b].w,
                     ti92_buttons[b].h, 
                     ti92_xor_color[ti92_buttons[b].xor_id], PG_XOR);
  }
}

static void
ti92_hide_cursor_button(void)
{
  int b = ti92_cursor_button;
  if (ti92_cursor_button_on) {
    ti92_cursor_button_on = 0;
    pgFillRectangle( ti92_buttons[b].x, 
                     ti92_buttons[b].y,
                     ti92_buttons[b].w,
                     ti92_buttons[b].h, 
                     ti92_xor_color[ti92_buttons[b].xor_id], PG_XOR);
  }
}

static void
ti92_move_cursor_button_right(void)
{
  if ((ti92_cursor_button >= TI92_BUTTON_LN) && 
      (ti92_cursor_button <  TI92_BUTTON_ENTER4)) {
    ti92_hide_cursor_button();
    ti92_cursor_button++;
  } else
  if (ti92_cursor_button == TI92_BUTTON_ENTER4) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_LN;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_SIN) && 
      (ti92_cursor_button <  TI92_BUTTON_POWER)) {
    ti92_hide_cursor_button();
    ti92_cursor_button++;
  } else
  if (ti92_cursor_button == TI92_BUTTON_POWER) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_SIN;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_PALEFT) && 
      (ti92_cursor_button <  TI92_BUTTON_DIVIDE)) {
    ti92_hide_cursor_button();
    ti92_cursor_button++;
  } else
  if (ti92_cursor_button == TI92_BUTTON_DIVIDE) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_PALEFT;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_Q) && 
      (ti92_cursor_button <  TI92_BUTTON_MULTIPLY)) {
    ti92_hide_cursor_button();
    ti92_cursor_button++;
  } else
  if (ti92_cursor_button == TI92_BUTTON_MULTIPLY) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_Q;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_A) && 
      (ti92_cursor_button <  TI92_BUTTON_MINUS)) {
    ti92_hide_cursor_button();
    ti92_cursor_button++;
  } else
  if (ti92_cursor_button == TI92_BUTTON_MINUS) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_A;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_SHIFT) && 
      (ti92_cursor_button <  TI92_BUTTON_PLUS)) {
    ti92_hide_cursor_button();
    ti92_cursor_button++;
  } else
  if (ti92_cursor_button == TI92_BUTTON_PLUS) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_SHIFT;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_ON) && 
      (ti92_cursor_button <  TI92_BUTTON_ENTER3)) {
    ti92_hide_cursor_button();
    ti92_cursor_button++;
  } else
  if (ti92_cursor_button == TI92_BUTTON_ENTER3) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_ON;
  }
}

static void
ti92_move_cursor_button_left(void)
{
  if ((ti92_cursor_button >  TI92_BUTTON_LN) && 
      (ti92_cursor_button <= TI92_BUTTON_ENTER4)) {
    ti92_hide_cursor_button();
    ti92_cursor_button--;
  } else
  if (ti92_cursor_button == TI92_BUTTON_LN) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_ENTER4;
  } else 
  if ((ti92_cursor_button >  TI92_BUTTON_SIN) && 
      (ti92_cursor_button <= TI92_BUTTON_POWER)) {
    ti92_hide_cursor_button();
    ti92_cursor_button--;
  } else
  if (ti92_cursor_button == TI92_BUTTON_SIN) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_POWER;
  } else 
  if ((ti92_cursor_button >  TI92_BUTTON_PALEFT) && 
      (ti92_cursor_button <= TI92_BUTTON_DIVIDE)) {
    ti92_hide_cursor_button();
    ti92_cursor_button--;
  } else
  if (ti92_cursor_button == TI92_BUTTON_PALEFT) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_DIVIDE;
  } else 
  if ((ti92_cursor_button >  TI92_BUTTON_Q) && 
      (ti92_cursor_button <= TI92_BUTTON_MULTIPLY)) {
    ti92_hide_cursor_button();
    ti92_cursor_button--;
  } else
  if (ti92_cursor_button == TI92_BUTTON_Q) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_MULTIPLY;
  } else 
  if ((ti92_cursor_button >  TI92_BUTTON_A) && 
      (ti92_cursor_button <= TI92_BUTTON_MINUS)) {
    ti92_hide_cursor_button();
    ti92_cursor_button--;
  } else
  if (ti92_cursor_button == TI92_BUTTON_A) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_MINUS;
  } else 
  if ((ti92_cursor_button >  TI92_BUTTON_SHIFT) && 
      (ti92_cursor_button <= TI92_BUTTON_PLUS)) {
    ti92_hide_cursor_button();
    ti92_cursor_button--;
  } else
  if (ti92_cursor_button == TI92_BUTTON_SHIFT) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_PLUS;
  } else 
  if ((ti92_cursor_button >  TI92_BUTTON_ON) && 
      (ti92_cursor_button <= TI92_BUTTON_ENTER3)) {
    ti92_hide_cursor_button();
    ti92_cursor_button--;
  } else
  if (ti92_cursor_button == TI92_BUTTON_ON) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_ENTER3;
  }
}

static void
ti92_move_cursor_button_up(void)
{
  if (ti92_cursor_button == TI92_BUTTON_2ND_2) {
    /* Nothing to do */
  } else 
  if (ti92_cursor_button == TI92_BUTTON_MODE) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_2ND_2;
  } else
  if (ti92_cursor_button == TI92_BUTTON_CLEAR) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_MODE;
  } else
  if (ti92_cursor_button == TI92_BUTTON_LN) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_CLEAR;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_SIN) && 
      (ti92_cursor_button <= TI92_BUTTON_POWER)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_SIN) + TI92_BUTTON_LN;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_PALEFT) && 
      (ti92_cursor_button <= TI92_BUTTON_DIVIDE)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_PALEFT) + TI92_BUTTON_SIN;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_7) && 
      (ti92_cursor_button <= TI92_BUTTON_MULTIPLY)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_7) + TI92_BUTTON_PALEFT;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_4) && 
      (ti92_cursor_button <= TI92_BUTTON_MINUS)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_4) + TI92_BUTTON_7;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_1) && 
      (ti92_cursor_button <= TI92_BUTTON_PLUS)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_1) + TI92_BUTTON_4;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_0) && 
      (ti92_cursor_button <= TI92_BUTTON_ENTER3)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_0) + TI92_BUTTON_1;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_Q) && 
      (ti92_cursor_button <= TI92_BUTTON_P)) {
    /* Nothing to do */
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_A) && 
      (ti92_cursor_button <= TI92_BUTTON_L)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_A) + TI92_BUTTON_Q;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_SHIFT) && 
      (ti92_cursor_button <= TI92_BUTTON_ENTER_1)) {
    ti92_hide_cursor_button();
    if (ti92_cursor_button == TI92_BUTTON_ENTER_1) ti92_cursor_button = TI92_BUTTON_L;
    else ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_SHIFT) + TI92_BUTTON_A;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_ON) && 
      (ti92_cursor_button <= TI92_BUTTON_SPACE)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_ON) + TI92_BUTTON_SHIFT;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_EQUALS) && 
      (ti92_cursor_button <= TI92_BUTTON_ENTER_2)) {
    ti92_hide_cursor_button();
    if (ti92_cursor_button == TI92_BUTTON_ENTER_2) ti92_cursor_button = TI92_BUTTON_L;
    else ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_EQUALS) + TI92_BUTTON_M;
  }
}

static void
ti92_move_cursor_button_down(void)
{
  if (ti92_cursor_button == TI92_BUTTON_2ND_2) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_MODE;
  } else 
  if (ti92_cursor_button == TI92_BUTTON_MODE) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_CLEAR;
  } else
  if (ti92_cursor_button == TI92_BUTTON_CLEAR) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_LN;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_LN) && 
      (ti92_cursor_button <= TI92_BUTTON_ENTER4)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_LN) + TI92_BUTTON_SIN;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_SIN) && 
      (ti92_cursor_button <= TI92_BUTTON_POWER)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_SIN) + TI92_BUTTON_PALEFT;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_PALEFT) && 
      (ti92_cursor_button <= TI92_BUTTON_DIVIDE)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_PALEFT) + TI92_BUTTON_7;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_7) && 
      (ti92_cursor_button <= TI92_BUTTON_MULTIPLY)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_7) + TI92_BUTTON_4;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_4) && 
      (ti92_cursor_button <= TI92_BUTTON_MINUS)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_4) + TI92_BUTTON_1;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_1) && 
      (ti92_cursor_button <= TI92_BUTTON_PLUS)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_1) + TI92_BUTTON_0;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_0) && 
      (ti92_cursor_button <= TI92_BUTTON_ENTER3)) {
    /* Nothing to do */
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_Q) && 
      (ti92_cursor_button <= TI92_BUTTON_P)) {
    ti92_hide_cursor_button();
    if (ti92_cursor_button == TI92_BUTTON_P) ti92_cursor_button = TI92_BUTTON_L;
    else ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_Q) + TI92_BUTTON_A;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_A) && 
      (ti92_cursor_button <= TI92_BUTTON_L)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_A) + TI92_BUTTON_SHIFT;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_SHIFT) && 
      (ti92_cursor_button <= TI92_BUTTON_C)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_SHIFT) + TI92_BUTTON_ON;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_V) && 
      (ti92_cursor_button <= TI92_BUTTON_N)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = TI92_BUTTON_SPACE;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_M) && 
      (ti92_cursor_button <= TI92_BUTTON_ENTER_1)) {
    ti92_hide_cursor_button();
    ti92_cursor_button = (ti92_cursor_button - TI92_BUTTON_M) + TI92_BUTTON_EQUALS;
  } else
  if ((ti92_cursor_button >= TI92_BUTTON_ON) && 
      (ti92_cursor_button <= TI92_BUTTON_ENTER_2)) {
    /* Nothing to do */
  }
}

void 
kbd_get_analog_direction(int Analog_x, int Analog_y, int *x, int *y)
{
  int DeltaX = 255;
  int DeltaY = 255;
  int DirX   = 0;
  int DirY   = 0;

  *x = 0;
  *y = 0;

  if (Analog_x <=  55) { DeltaX = Analog_x; DirX = -1; }
  else 
  if (Analog_x >= 200) { DeltaX = 255 - Analog_x; DirX = 1; }

  if (Analog_y <=  55) { DeltaY = Analog_y; DirY = -1; }
  else 
  if (Analog_y >= 200) { DeltaY = 255 - Analog_y; DirY = 1; }

  if (DeltaX < DeltaY) *x = DirX;
  else                 *y = DirY;
}

int
kbd_scan_keyboard(void)
{
  SceCtrlData c;
  int         event;
  int         b;

  int new_Lx;
  int new_Ly;
  int old_Lx;
  int old_Ly;

# if 0 
  pspDebugPrintf("TimeStamp=%d Buttons=%x Lx=%d Ly=%d\n",
      ctrl.TimeStamp, ctrl.Buttons, ctrl.Lx, ctrl.Ly);
# endif
  event = 0;
  sceCtrlPeekBufferPositive( &c, 1 );
  // sceCtrlReadBufferPositive(&c, 1);

  if ((c.Buttons & (PSP_CTRL_LTRIGGER|PSP_CTRL_RTRIGGER|PSP_CTRL_START)) ==
       (PSP_CTRL_LTRIGGER|PSP_CTRL_RTRIGGER|PSP_CTRL_START)) {
    /* Exit ! */
    psp_exit(0);
  } else 
  if ((c.Buttons & (PSP_CTRL_LTRIGGER|PSP_CTRL_START|PSP_CTRL_SELECT)) == 
           (PSP_CTRL_LTRIGGER|PSP_CTRL_START|PSP_CTRL_SELECT)) {
    /* Save ram */
    save_emulator_stuff();
  }

  ti92_show_cursor_button();
  
  /* Check Analog Device */
  kbd_get_analog_direction(loc_button_data.Lx,loc_button_data.Ly,&old_Lx,&old_Ly);
  kbd_get_analog_direction( c.Lx, c.Ly, &new_Lx, &new_Ly);

  if ((c.TimeStamp - loc_last_analog_time) > KBD_MIN_ANALOG_TIME) {
    loc_last_analog_time = c.TimeStamp;
    /* Analog device has moved */
    if (new_Lx > 0) {
      ti92_move_cursor_button_right();
    } else 
    if (new_Lx < 0) {
      ti92_move_cursor_button_left();
    } else
    if (new_Ly < 0) {
      ti92_move_cursor_button_up();
    } else 
    if (new_Ly > 0) {
      ti92_move_cursor_button_down();
    }
  }

  for (b = 0; b < KBD_MAX_BUTTONS; b++) {
    if (c.Buttons & loc_button_mask[b]) {
      if (!(loc_button_data.Buttons & loc_button_mask[b])) {
        loc_button_press[b] = 1;
        event = 1;
      }
    } else {
      if (loc_button_data.Buttons & loc_button_mask[b]) {
        loc_button_release[b] = 1;
        event = 1;
      }
    }
  }
  memcpy(&loc_button_data,&c,sizeof(SceCtrlData));

  return event;
}

void
kbd_wait_start(void)
{
  while (1)
  {
    SceCtrlData c;
    sceCtrlReadBufferPositive(&c, 1);
    if (c.Buttons & PSP_CTRL_START) break;
  }
}

void
psp_init_keyboard(void)
{
  int i;
  for(i=0; i<120; i++) {
    keyStates[i] = 0;
  }
}

int
psp_is_key_pressed(int key)
{
  return keyStates[key];
}

int
psp_update_keys(void)
{
  int         b;

  static char first_time = 1;
  static long first_time_stamp = -1;
  static int release_pending = 0;
  static int release_button  = 0;

  if (first_time) {

    SceCtrlData c;
    sceCtrlPeekBufferPositive(&c, 1);
    if (first_time_stamp == -1) first_time_stamp = c.TimeStamp;
    if ((! c.Buttons) && ((c.TimeStamp - first_time_stamp) < KBD_MIN_START_TIME)) return 0;
    
    psp_display_screen_on = 1;
    first_time = 0;

    release_pending = 0;
    release_button  = TI92_BUTTON_Q;

    for (b = 0; b < KBD_MAX_BUTTONS; b++) {
      loc_button_release[b] = 0;
      loc_button_press[b] = 0;
    }
    sceCtrlPeekBufferPositive(&loc_button_data, 1);
    return 0;
  }

  if (release_pending)
  {
    release_pending = 0;
    return ti92_decode_key(release_button,0);
  }

  if (kbd_scan_keyboard()) {

    /* check release event */
    for (b = 0; b < KBD_MAX_BUTTONS; b++) {
      if (loc_button_release[b]) {
        loc_button_release[b] = 0;
        release_button        = b;
        release_pending       = 1;
        break;
      } 
    }
    /* check press event */
    for (b = 0; b < KBD_MAX_BUTTONS; b++) {
      if (loc_button_press[b]) {
        loc_button_press[b] = 0;
        release_pending     = 0;
        return ti92_decode_key(b, 1);
      }
    }
  }
  return 0;
}
