/*
 *  xtiger port on PSP 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

 /*
  */

#include <stdlib.h>
#include "sysdeps.h"
#include <stdio.h>
#include <assert.h>
#include <ctype.h>

#include <dirent.h>
#include <sys/stat.h>
#include <string.h>
 
#include "config.h"
#include "options.h"
#include "memory.h"
#include "newcpu.h"
#include "debug.h"
#include "hardware.h"
#include "specific.h"
#include "globinfo.h"
#include "global.h"

#include "psp_fmgr.h"

struct GlobalInformation globInf;

/*
CPTR nextpc;

int enterDebugger = 0;
*/

int breakPoints = 0;
ULONG breakAdr[16];
UWORD breakSave[16];
int activeBreak = -1;


int intlink = 0;
int romsize = 0;
int dontSaveState = 1;

#define DEFAULT_CFGFILE "/tiger.cfg"
#define DEFAULT_ROMFILE "/ti92.rom"
#define DEFAULT_MEMFILE "/ti92.mem"

#define DEFAULT_DEVFILE "/dev/ti"


int
save_emulator_stuff(void)
{
  FILE *fp;

  if(!dontSaveState && strlen(globInf.memFile) && (fp = fopen(globInf.memFile, "wb"))) {
    int m = MEM_SIZE;
    m68k_setpc(m68k_getpc());
    MakeSR();
    fwrite(&m, 1, sizeof(int), fp);
    fwrite(ti_mem, 1, MEM_SIZE, fp);
    fwrite(ti_io, 1, IO_SIZE, fp);
    fwrite(&regs, 1, sizeof(regs), fp); 
    fwrite(&timerVal, 1, sizeof(timerVal), fp);
    fwrite(&specialflags, 1, sizeof(specialflags), fp);
    fclose(fp); 
  }

  return 0;

# if 0
  exit_hardware();
  screen_off();
  exit_specific();
  mem_exit();
# endif
}

int 
load_prog_file()
{
  static char prog_filedir[PSP_FMGR_MAX_NAME];
  static int  first = 1;

  char prog_filename[PSP_FMGR_MAX_NAME];
  struct stat    aStat;
  UBYTE         *backBuf;

  if (first) {
    first = 0;
    strcpy(prog_filedir, homeDirectory);
    strcat(prog_filedir, "/");
  }

  if (psp_file_request(prog_filename, prog_filedir)) {
    if (stat(prog_filename, &aStat)) return 0;
  
    backBuf = (UBYTE*)malloc(128*1024);
    send_ti_file(prog_filename,backBuf);
    free(backBuf);
  }

  return 1;
}

int 
run_emulator(void)
{
  FILE *fp;

  strcpy(globInf.cfgFile, homeDirectory);
  strcat(globInf.cfgFile, DEFAULT_CFGFILE);

  strcpy(globInf.romFile, homeDirectory);
  strcat(globInf.romFile, DEFAULT_ROMFILE);

  strcpy(globInf.memFile, homeDirectory);
  strcat(globInf.memFile, DEFAULT_MEMFILE);

  strcpy(globInf.devFile, homeDirectory);
  strcat(globInf.devFile, DEFAULT_DEVFILE);
  globInf.memSize = 256;
  globInf.grayPlanes = 2;
  cycleInstr = globInf.itick = 640;
  globInf.tickRate = 40000;
  globInf.cycleRate = 150;
  globInf.darkCol = 0x000034;
  globInf.lightCol = 0x989898;
  globInf.syncOne = 0;
  
  load_cfg_file(globInf.cfgFile);
  
  if((fp = fopen(globInf.romFile, "rb"))) {
    fseek(fp, 0, SEEK_END);
    romsize = ftell(fp);
    fclose(fp);
  }
  else {
    pspDebugScreenPrintf("**Error: Cant find \"%s\"!\n", globInf.romFile);
    psp_exit(10);
  }

  if(!(romsize == 1024*1024 || romsize == 1024*1024*2)) {
    pspDebugScreenPrintf("**Error: Illegal romsize. Must be 1Mb or 2Mb.\n");
    psp_exit(10);
  }
    
  if(!(mem_init(globInf.memSize, romsize))) {
    pspDebugScreenPrintf("**Error: Could not initialize memory.\n");
    psp_exit(10);
  }    
    
  init_m68k();
  
  if((fp = fopen(globInf.romFile, "rb"))) {
    fread(ti_rom, 1, romsize, fp);
    fclose(fp);
  }
  else
    psp_exit(100);

  init_keyboard();
  init_specific();
  screen_on();
  set_screen_ptr(ti_mem);

  init_hardware(globInf.devFile);

  if((fp = fopen(globInf.memFile, "rb"))) {
    int m;
    fread(&m, 1, sizeof(int), fp);
    if(m != MEM_SIZE) {
      fclose(fp);

      rom_at_0();
      MC68000_reset();
      ram_at_0();
      
    }
    else {
      fread(ti_mem, 1, MEM_SIZE, fp);
      fread(ti_io, 1, IO_SIZE, fp);
      fread(&regs, 1, sizeof(regs), fp);
      fread(&timerVal, 1, sizeof(timerVal), fp);
      fread(&specialflags, 1, sizeof(specialflags), fp);
      fclose(fp); 
      MakeFromSR();
      m68k_setpc(regs.pc);
      update_contrast();
      update_bitmap();
    }
  }
  else {
    rom_at_0();
    MC68000_reset();
    ram_at_0();
    
  }

#ifdef PENT_COUNTER
  if(globInf.tickRate)
    calibrate_pcounter();
#endif

  /*
   *  Create the TI window
   */
  psp_init_graphic();

  dontSaveState = 0;

  MC68000_run();

  save_emulator_stuff();

  return 0;
}
